#ifndef __AS_AUDIO
#define __AS_AUDIO
#endif


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include "forms.h"
#include "xwin2.h"
#include "global.h"
#include "error.h"
#include "audio.h"
#include "file.h"
#include "tools.h"


int InitSoundDevice(int Mode)
{
  if(_OpenSoundDevice(Mode)!=0)
    return -1;

#ifdef DEBUG
  fprintf(stderr,"MODE: %d\n",Mode);
#endif

/*
  if(SetupSoundDeviceSetBufferSize(0x7fff0010)!=0)
    ShowErrorMsg();    
*/
    
  if((Mode==PLAYANDRECTRACKS)&&(Opt->Audio.FullDuplex==YES))
  {
    if( (strcmp((char *)&Opt->Audio.DeviceNamePlay,(char *)&Opt->Audio.DeviceNameRec)!=0)||
        (Opt->Audio.DevicePlay != Opt->Audio.DeviceRec) )
    {
      StoreErrorMsg("full duplex error: device for play and rec not equal",ERR_CRITICAL);
      return -1;
    }

    if(_SetupSoundDeviceCheckFullDuplex()!=0)
      return -1;
    if(_SetupSoundDeviceSetFullDuplex()!=0)
      return -1;
    if(_SetupSoundDeviceSetBufferSizePlay(CalcFragInfo(Opt->Audio.DuplexPlayFragsize,Opt->Audio.DuplexPlayFragments))!=0)
      return -1;
    if(_SetupSoundDeviceFormatPlay( (Opt->Audio.Bits==8) ? BIT8 : BIT16 )!=0)
      return -1;
    if(_SetupSoundDeviceChannelsPlay(Opt->Audio.Channels)!=0)
      return -1;
    if(_SetupSoundDeviceSampleRatePlay((int)Opt->Audio.SampleRate)!=0)
      return -1;
    if(_SetupSoundDeviceGetFullDuplexInfo()!=0)
      return -1;
  }



  if((Mode==PLAYANDRECTRACKS)&&(Opt->Audio.FullDuplex==NO))
  {
    StoreErrorMsg("play and rec together currently only\nsupported with full duplex cards",ERR_CRITICAL);
      return -1;
  }


  if(Mode==PLAYTRACKS)
  {

//    if(_SetupSoundDeviceResetPlay()!=0)
//      return -1;
    if(_SetupSoundDeviceSetBufferSizePlay(CalcFragInfo(Opt->Audio.PlayFragsize,Opt->Audio.PlayFragments))!=0)
      return -1;
    if(_SetupSoundDeviceFormatPlay( (Opt->Audio.Bits==8) ? BIT8 : BIT16 )!=0)
      return -1;
    if(_SetupSoundDeviceChannelsPlay(Opt->Audio.Channels)!=0)
      return -1;
    if(_SetupSoundDeviceSampleRatePlay((int)Opt->Audio.SampleRate)!=0)
      return -1;
    if(_SetupSoundDeviceGetPlayInfo()!=0)
      return -1;
  }

  if(Mode==RECTRACKS)
  {
//    if(_SetupSoundDeviceResetRec()!=0)
//      return -1;
    if(_SetupSoundDeviceSetBufferSizeRec(CalcFragInfo(Opt->Audio.RecFragsize,Opt->Audio.RecFragments))!=0)
      return -1;
    if(_SetupSoundDeviceFormatRec( (Opt->Audio.Bits==8) ? BIT8 : BIT16 )!=0)
      return -1;
    if(_SetupSoundDeviceChannelsRec(Opt->Audio.Channels)!=0)
      return -1;
    if(_SetupSoundDeviceSampleRateRec((int)Opt->Audio.SampleRate)!=0)
      return -1;
    if(_SetupSoundDeviceGetRecInfo()!=0)
      return -1;
  }


  return 0;
}


int _OpenSoundDevice(int Mode)
{
  switch(Mode)
  {
    case PLAYTRACKS:
      return(_OpenSoundDeviceForPlay());
      break;

    case RECTRACKS:
      return(_OpenSoundDeviceForRec());
      break;

    case PLAYANDRECTRACKS:
      return(_OpenSoundDeviceForPlayAndRec());
      break;

    default:
      return -1;
  }
}


int CloseSoundDevice()
{
  int RetCode=0;

  if(Opt->Audio.DevicePlay==Opt->Audio.DeviceRec)
  {
    if(close(Opt->Audio.DevicePlay)!=0)
      RetCode--;
    else
    {
      Opt->Audio.DevicePlay=0;
      Opt->Audio.DeviceRec=0;
    }
  }

  if(Opt->Audio.DevicePlay!=0)
  {
    if(close(Opt->Audio.DevicePlay)!=0)
      RetCode--;
    else
      Opt->Audio.DevicePlay=0;
  }

  if(Opt->Audio.DeviceRec!=0)
  {
    if(close(Opt->Audio.DeviceRec)!=0)
      RetCode-=2;
    else
      Opt->Audio.DeviceRec=0;
  }
    

  switch(RetCode)
  {
    case -1: StoreErrorMsg("unable to close play sounddevice",ERR_PANIC); break;
    case -2: StoreErrorMsg("unable to close rec sounddevice",ERR_PANIC); break;
    case -3: StoreErrorMsg("unable to close play and rec sounddevice",ERR_PANIC); break;
  }

  return RetCode;
}


int FlushSoundDevice()
{
  return(_SetupSoundDeviceFlushAudioPlay());
}



/**************************************************************************************
 **************************************************************************************
 **************************************************************************************/

int GetBestPlayInfo()
{
  Opt->Audio.PlayInfoTested=NO;

  if(_OpenSoundDevice(PLAYTRACKS)!=0)
    return -1;

  if(_SetupSoundDeviceFormatPlay( (Opt->Audio.Bits==8) ? BIT8 : BIT16 )!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  if(_SetupSoundDeviceChannelsPlay(Opt->Audio.Channels)!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  if(_SetupSoundDeviceSampleRatePlay((int)Opt->Audio.SampleRate)!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  if(_SetupSoundDeviceGetPlayInfo()!=0)
  {
    CloseSoundDevice();
    return -1;
  }

  CloseSoundDevice();
  return 0;
}

int TestPlayInfo()
{
  long FragInfo=0;
  
  if(_OpenSoundDevice(PLAYTRACKS)!=0)
    return -1;

  /* best match */
  FragInfo=CalcFragInfo(Opt->Audio.PlayFragsize,Opt->Audio.PlayFragments);


  if(_SetupSoundDeviceSetBufferSizePlay(FragInfo)!=0)
  {
    CloseSoundDevice();
    return -1;
  }

  if(_SetupSoundDeviceFormatPlay( (Opt->Audio.Bits==8) ? BIT8 : BIT16 )!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  if(_SetupSoundDeviceChannelsPlay(Opt->Audio.Channels)!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  if(_SetupSoundDeviceSampleRatePlay((int)Opt->Audio.SampleRate)!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  if(_SetupSoundDeviceGetPlayInfo()!=0)
  {
    CloseSoundDevice();
    return -1;
  }

  CloseSoundDevice();

  Opt->Audio.PlayInfoTested=YES;

  return 0;
}


int GetBestRecInfo()
{
  Opt->Audio.RecInfoTested=NO;

  if(_OpenSoundDevice(RECTRACKS)!=0)
    return -1;

  if(_SetupSoundDeviceFormatRec( (Opt->Audio.Bits==8) ? BIT8 : BIT16 )!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  if(_SetupSoundDeviceChannelsRec(Opt->Audio.Channels)!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  if(_SetupSoundDeviceSampleRateRec((int)Opt->Audio.SampleRate)!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  if(_SetupSoundDeviceGetRecInfo()!=0)
  {
    CloseSoundDevice();
    return -1;
  }

  CloseSoundDevice();
  return 0;

}



int TestRecInfo()
{
  long FragInfo=0;
  
  if(_OpenSoundDevice(RECTRACKS)!=0)
    return -1;

  /* best match */
  FragInfo=CalcFragInfo(Opt->Audio.RecFragsize,Opt->Audio.RecFragments);



  if(_SetupSoundDeviceSetBufferSizeRec(FragInfo)!=0)
  {
    CloseSoundDevice();
    return -1;
  }

  if(_SetupSoundDeviceFormatRec( (Opt->Audio.Bits==8) ? BIT8 : BIT16 )!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  if(_SetupSoundDeviceChannelsRec(Opt->Audio.Channels)!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  if(_SetupSoundDeviceSampleRateRec((int)Opt->Audio.SampleRate)!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  if(_SetupSoundDeviceGetRecInfo()!=0)
  {
    CloseSoundDevice();
    return -1;
  }

  CloseSoundDevice();

  Opt->Audio.RecInfoTested=YES;

  return 0;
}




int GetBestDuplexInfo()
{

  Opt->Audio.DuplexInfoTested=NO;


  if(strcmp((char *)&Opt->Audio.DeviceNamePlay,(char *)&Opt->Audio.DeviceNameRec)!=0)
  {
    StoreErrorMsg("full duplex error: device for play and rec not equal",ERR_CRITICAL);
    return -1;
  }

  if(_OpenSoundDevice(PLAYANDRECTRACKS)!=0)
  {
    CloseSoundDevice();
    return -1;
  }

  if(_SetupSoundDeviceCheckFullDuplex()!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  if(_SetupSoundDeviceSetFullDuplex()!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  if(_SetupSoundDeviceFormatPlay( (Opt->Audio.Bits==8) ? BIT8 : BIT16 )!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  if(_SetupSoundDeviceChannelsPlay(Opt->Audio.Channels)!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  if(_SetupSoundDeviceSampleRatePlay((int)Opt->Audio.SampleRate)!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  if(_SetupSoundDeviceGetFullDuplexInfo()!=0)
  {
    CloseSoundDevice();
    return -1;
  }

  CloseSoundDevice();

  if(Opt->Audio.DuplexPlayFragsize != Opt->Audio.DuplexRecFragsize)
  {
    StoreErrorMsg("play and rec fragsize must be equal in size!",ERR_CRITICAL);
    return -1;
  }

  if(Opt->Audio.DuplexPlayFragments != Opt->Audio.DuplexRecFragments)
  {
    StoreErrorMsg("play and rec fragments must be equal in size!",ERR_CRITICAL);
    return -1;
  }

  return 0;

}



int TestDuplexInfo()
{
  long FragInfo=0;
  short Fragments=0;
  short Fragsize=0;

  if(strcmp((char *)&Opt->Audio.DeviceNamePlay,(char *)&Opt->Audio.DeviceNameRec)!=0)
  {
    StoreErrorMsg("full duplex error: device for play and rec not equal",ERR_CRITICAL);
    return -1;
  }
  
  if(_OpenSoundDevice(PLAYANDRECTRACKS)!=0)
    return -1;

  /* use lowest values */
  if(Opt->Audio.DuplexRecFragsize <= Opt->Audio.DuplexPlayFragsize)
    Fragsize=Opt->Audio.DuplexRecFragsize;
  else
    Fragsize=Opt->Audio.DuplexPlayFragsize;
  
  if(Opt->Audio.DuplexRecFragments <= Opt->Audio.DuplexPlayFragments)
    Fragments=Opt->Audio.DuplexRecFragments;
  else
    Fragments=Opt->Audio.DuplexPlayFragments;
      
  FragInfo=CalcFragInfo(Fragsize,Fragments);


  if(_SetupSoundDeviceCheckFullDuplex()!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  if(_SetupSoundDeviceSetFullDuplex()!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  /* this works for in AND output */
  if(_SetupSoundDeviceSetBufferSizePlay(FragInfo)!=0)
  {
    CloseSoundDevice();
    return -1;
  }

  if(_SetupSoundDeviceFormatPlay( (Opt->Audio.Bits==8) ? BIT8 : BIT16 )!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  if(_SetupSoundDeviceChannelsPlay(Opt->Audio.Channels)!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  if(_SetupSoundDeviceSampleRatePlay((int)Opt->Audio.SampleRate)!=0)
  {
    CloseSoundDevice();
    return -1;
  }
  if(_SetupSoundDeviceGetFullDuplexInfo()!=0)
  {
    CloseSoundDevice();
    return -1;
  }

  CloseSoundDevice();

  if(Opt->Audio.DuplexPlayFragsize != Opt->Audio.DuplexRecFragsize)
  {
    StoreErrorMsg("play and rec fragsize must be equal in size!",ERR_CRITICAL);
    return -1;
  }
  if(Opt->Audio.DuplexPlayFragments != Opt->Audio.DuplexRecFragments)
  {
    StoreErrorMsg("play and rec fragments must be equal in size!",ERR_CRITICAL);
    return -1;
  }

  Opt->Audio.DuplexInfoTested=YES;

  return 0;
}



/**************************************************************************************
 **************************************************************************************
 **************************************************************************************/

int _OpenSoundDeviceForPlay()
{
  int i;
  i=open( (const char *)&Opt->Audio.DeviceNamePlay,O_WRONLY);
  if(i<0)
  {
    StoreErrorMsg("unable to open sounddevice in playmode",ERR_PANIC);
    return -1;
  }
  Opt->Audio.DevicePlay=i;
  return 0;
}

int _OpenSoundDeviceForRec()
{
  int i;
  i=open( (const char *)&Opt->Audio.DeviceNameRec,O_RDONLY);
  if(i<0)
  {
    StoreErrorMsg("unable to open sounddevice in recmode",ERR_PANIC);
    return -1;
  }
  Opt->Audio.DeviceRec=i;
  return 0;
}

int _OpenSoundDeviceForPlayAndRec()
{
  int i;
  i=open( (const char *)&Opt->Audio.DeviceNameRec,O_RDWR);
  if(i<0)
  {
    StoreErrorMsg("unable to open sounddevice in playrecmode",ERR_PANIC);
    return -1;
  }
  Opt->Audio.DevicePlay=i;
  Opt->Audio.DeviceRec=i;
  return 0;
}

/**************************************************************************************
 **************************************************************************************
 **************************************************************************************/


int _SetupSoundDeviceResetPlay()
{
   if(ioctl(Opt->Audio.DevicePlay,SNDCTL_DSP_RESET)!=0)
   {
     StoreErrorMsg("unable to reset play sounddevice",ERR_WARNING);
     return -1;
   }
   return 0;
}

int _SetupSoundDeviceFlushAudioPlay()
{
   if(ioctl(Opt->Audio.DevicePlay,SNDCTL_DSP_SYNC,0)!=0)
   {
     StoreErrorMsg("unable to flush play audio",ERR_WARNING);
     return -1;
   }
   return 0;
}


int _SetupSoundDeviceFormatPlay(int Format)
{
   if(ioctl(Opt->Audio.DevicePlay,SNDCTL_DSP_SETFMT,&Format)!=0)
   {
     StoreErrorMsg("unable to set play bits",ERR_WARNING);
     return -1;
   }
   return 0;
}

int _SetupSoundDeviceChannelsPlay(int Channels)
{
   if(ioctl(Opt->Audio.DevicePlay,SNDCTL_DSP_CHANNELS,&Channels)!=0)
   {
     StoreErrorMsg("unable to set play channels",ERR_WARNING);
     return -1;
   }
   return 0;
}


int _SetupSoundDeviceSampleRatePlay(int SampleRate)
{
   if(ioctl(Opt->Audio.DevicePlay,SNDCTL_DSP_SPEED,&SampleRate)!=0)
   {
     StoreErrorMsg("unable to set play sample rate",ERR_WARNING);
     return -1;
   }
   return 0;
}

int _SetupSoundDeviceSetBufferSizePlay(int Fragments)
{
   /* DANGEROUS: this call does not always return !=0 if the buffer
    * size couldn't be set!
    * We have to verify the settings later ! */
   if(ioctl(Opt->Audio.DevicePlay,SNDCTL_DSP_SETFRAGMENT,&Fragments)!=0)
   {
     StoreErrorMsg("unable to set play buffer size",ERR_WARNING);
     return -1;
   }
   return 0;
}

/**************************************************************************************
 **************************************************************************************
 **************************************************************************************/

int _SetupSoundDeviceResetRec()
{
   if(ioctl(Opt->Audio.DeviceRec,SNDCTL_DSP_RESET)!=0)
   {
     StoreErrorMsg("unable to reset rec sounddevice",ERR_WARNING);
     return -1;
   }
   return 0;
}


int _SetupSoundDeviceFormatRec(int Format)
{

   if(ioctl(Opt->Audio.DeviceRec,SNDCTL_DSP_SETFMT,&Format)!=0)
   {
     StoreErrorMsg("unable to set rec bits",ERR_WARNING);
     return -1;
   }

   return 0;
}

int _SetupSoundDeviceChannelsRec(int Channels)
{
   if(ioctl(Opt->Audio.DeviceRec,SNDCTL_DSP_CHANNELS,&Channels)!=0)
   {
     StoreErrorMsg("unable to set rec channels",ERR_WARNING);
     return -1;
   }
   return 0;
}


int _SetupSoundDeviceSampleRateRec(int SampleRate)
{
   if(ioctl(Opt->Audio.DeviceRec,SNDCTL_DSP_SPEED,&SampleRate)!=0)
   {
     StoreErrorMsg("unable to set rec sample rate",ERR_WARNING);
     return -1;
   }
   return 0;
}

int _SetupSoundDeviceSetBufferSizeRec(int Fragments)
{
   /* DANGEROUS: this call does not always return !=0 if the buffer
    * size couldn't be set!
    * We have to verify the settings later ! */
   if(ioctl(Opt->Audio.DeviceRec,SNDCTL_DSP_SETFRAGMENT,&Fragments)!=0)
   {
     StoreErrorMsg("unable to set rec buffer size",ERR_WARNING);
     return -1;
   }
   return 0;
}



/**************************************************************************************
 **************************************************************************************
 **************************************************************************************/


int _SetupSoundDeviceSetFullDuplex()
{
   if(Opt->Audio.DevicePlay!=Opt->Audio.DeviceRec)
     return -1;

   if(ioctl(Opt->Audio.DevicePlay,SNDCTL_DSP_SETDUPLEX,0)!=0)
   {
     StoreErrorMsg("unable to set full duplex mode",ERR_WARNING);
     return -1;
   }
   return 0;
}


int _SetupSoundDeviceCheckFullDuplex()
{
  int Caps;



  if(ioctl(Opt->Audio.DevicePlay,SNDCTL_DSP_GETCAPS,&Caps)!=0)
  {
    StoreErrorMsg("can't read full duplex capabilities",ERR_CRITICAL);
  }

  if((Caps & DSP_CAP_DUPLEX)!=0)
    return 0;

  StoreErrorMsg("card does not support full duplex",ERR_CRITICAL);
  return -1;

}

int _SetupSoundDeviceGetPlayInfo()
{
  audio_buf_info PlayInfo;

  if(ioctl(Opt->Audio.DevicePlay,SNDCTL_DSP_GETOSPACE,&PlayInfo)!=0)
  {
    StoreErrorMsg("unable to get buffer size",ERR_WARNING);
    return -1;
  }
  Opt->Audio.PlayFragments=PlayInfo.fragments;
  Opt->Audio.PlayFragstotal=PlayInfo.fragstotal;
  Opt->Audio.PlayFragsize=PlayInfo.fragsize;
  Opt->Audio.PlayBytes=PlayInfo.bytes;

#ifdef DEBUG
  fprintf(stderr,"PLAY FRAGMENTS  : %d\n",PlayInfo.fragments);   
  fprintf(stderr,"PLAY FRAGSTOTAL : %d\n",PlayInfo.fragstotal);   
  fprintf(stderr,"PLAY FRAGSIZE   : %d\n",PlayInfo.fragsize);   
  fprintf(stderr,"PLAY BYTES      : %d\n",PlayInfo.bytes);   
#endif
   
  return 0;
}


int _SetupSoundDeviceGetRecInfo()
{
  audio_buf_info PlayInfo;

  if(ioctl(Opt->Audio.DeviceRec,SNDCTL_DSP_GETISPACE,&PlayInfo)!=0)
  {
    StoreErrorMsg("unable to get buffer size",ERR_WARNING);
    return -1;
  }
  Opt->Audio.RecFragments=PlayInfo.fragments;
  Opt->Audio.RecFragstotal=PlayInfo.fragstotal;
  Opt->Audio.RecFragsize=PlayInfo.fragsize;
  Opt->Audio.RecBytes=PlayInfo.bytes;

#ifdef DEBUG
  fprintf(stderr,"REC FRAGMENTS  : %d\n",PlayInfo.fragments);   
  fprintf(stderr,"REC FRAGSTOTAL : %d\n",PlayInfo.fragstotal);   
  fprintf(stderr,"REC FRAGSIZE   : %d\n",PlayInfo.fragsize);   
  fprintf(stderr,"REC BYTES      : %d\n",PlayInfo.bytes);   
#endif
   
  return 0;
}



int _SetupSoundDeviceGetFullDuplexInfo()
{
  audio_buf_info PlayInfo,RecInfo;

  if(ioctl(Opt->Audio.DevicePlay,SNDCTL_DSP_GETOSPACE,&PlayInfo)!=0)
  {
    StoreErrorMsg("unable to get full-duplex play buffer size",ERR_WARNING);
    return -1;
  }

  if(ioctl(Opt->Audio.DeviceRec,SNDCTL_DSP_GETISPACE,&RecInfo)!=0)
  {
    StoreErrorMsg("unable to get full-duplex rec size",ERR_WARNING);
    return -1;
  }


  Opt->Audio.DuplexPlayFragments=PlayInfo.fragments;
  Opt->Audio.DuplexPlayFragsize=PlayInfo.fragsize;
  Opt->Audio.DuplexPlayFragstotal=PlayInfo.fragstotal;
  Opt->Audio.DuplexPlayBytes=PlayInfo.bytes;
  Opt->Audio.DuplexRecFragments=RecInfo.fragments;
  Opt->Audio.DuplexRecFragsize=RecInfo.fragsize;
  Opt->Audio.DuplexRecFragstotal=RecInfo.fragstotal;
  Opt->Audio.DuplexRecBytes=RecInfo.bytes;
  
#ifdef DEBUG
  fprintf(stderr,"PLAY FRAGMENTS     : %d\n",PlayInfo.fragments);   
  fprintf(stderr,"PLAY FRAGSTOTAL    : %d\n",PlayInfo.fragstotal);   
  fprintf(stderr,"PLAY FRAGSIZE      : %d\n",PlayInfo.fragsize);   
  fprintf(stderr,"PLAY BYTES         : %d\n",PlayInfo.bytes);   
  fprintf(stderr,"REC FRAGMENTS      : %d\n",RecInfo.fragments);   
  fprintf(stderr,"REC FRAGSTOTAL     : %d\n",RecInfo.fragstotal);   
  fprintf(stderr,"REC FRAGSIZE       : %d\n",RecInfo.fragsize);   
  fprintf(stderr,"REC BYTES          : %d\n",RecInfo.bytes);   
#endif

  return 0;
}



/**************************************************************************************
 **************************************************************************************
 **************************************************************************************/



inline int PlayBufferToDevice (unsigned char *Buffer, int Size)
{
  write(Opt->Audio.DevicePlay,(unsigned char *)Buffer,Size);
  return 0;
}

inline int RecBufferFromDevice (unsigned char *Buffer, int Size)
{
  read(Opt->Audio.DeviceRec,(unsigned char *)Buffer,Size);
  return 0;
}



