#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "forms.h"
#include "xwin2.h"
#include "global.h"
#include "audio.h"
#include "error.h"
#include "tools.h"
#include "windows.h"
#include "tracks.h"

  int _ReadSettingsOutOfFile(FILE *CfgFile);
  void _WriteSettingsToFile(FILE *CfgFile);
  int CreateTrackPeakFile(struct tAudioTrack *Track, long SampleSkip);
  extern int LoadPeakFileToMemory(struct tAudioTrack *Track);
  


#define CFG_FILE_ITEMS  12

/******************************************************
 *** ReadConfigFile:
 *** reads the file ~/.XhdREC/XhdREC.cfg and sets all
 *** internal vars
 ******************************************************/

int ReadConfigFile()
{
  char tmp[300];
  FILE *CfgFile;
  int i;
  
  sprintf((char *)&tmp,"%s/.XhdREC.cfg",Opt->Prg.HomeDir);
  
  if( (CfgFile=fopen((char *)&tmp,"r"))==NULL )
  {
    StoreErrorMsg("config file not found",ERR_WARNING);
    return -1;
  }

  i=_ReadSettingsOutOfFile(CfgFile);
  fclose(CfgFile);

  if(i!=0)
  {
    StoreErrorMsg("config file corrupt",ERR_WARNING);
    return -1;
  }
  
  return 0;
}





int _ReadSettingsOutOfFile(FILE *CfgFile)
{
  char *ptmp;
  char tmp[300];
  int Items=0;     /* counts number of recognized items in cfg-file */
  
    
  /* parsing is bad code, but easy to read and change... */
  while(!feof(CfgFile))
  {
    fgets((char *)&tmp,299,CfgFile);

    /* AUDIO DIR */
    if((ptmp=strstr((char *)&tmp,"AUDIO_DIR="))!=NULL)
    {
      ptmp+=10;
      Items++;
      sprintf((char *)&Opt->Prg.AudioDir,"%s",ptmp);
      CutLF((char *)&Opt->Prg.AudioDir,299);
      continue;
    }

    /* MIXER APPLICATION */
    if((ptmp=strstr((char *)&tmp,"MIXER_APP="))!=NULL)
    {
      ptmp+=10;
      Items++;
      sprintf((char *)&Opt->Prg.MixerApp,"%s",ptmp);
      CutLF((char *)&Opt->Prg.MixerApp,299);
      continue;
    }

    /* HIDE WINDOWS WHILE RECORDING */
    if((ptmp=strstr((char *)&tmp,"HIDE_WINS_WHILE_REC="))!=NULL)
    {
      ptmp+=20;
      Items++;
      if(strstr(ptmp,"YES")!=NULL)
        Opt->Prg.HideWins=YES;
      else
        Opt->Prg.HideWins=NO;
      continue;
    }

    /* RECORD COUNTDOWN */
    if((ptmp=strstr((char *)&tmp,"REC_COUNTDOWN="))!=NULL)
    {
      ptmp+=14;
      Items++;
      Opt->Prg.RecCountdown=atoi(ptmp);
      continue;
    }

    /* PLAY DEVICE */
    if((ptmp=strstr((char *)&tmp,"PLAY_DEV="))!=NULL)
    {
      ptmp+=9;
      Items++;
      sprintf((char *)&Opt->Audio.DeviceNamePlay,"%s",ptmp);
      CutLF((char *)&Opt->Audio.DeviceNamePlay,299);
      continue;
    }

    /* REC DEVICE */
    if((ptmp=strstr((char *)&tmp,"REC_DEV="))!=NULL)
    {
      ptmp+=8;
      Items++;
      sprintf((char *)&Opt->Audio.DeviceNameRec,"%s",ptmp);
      CutLF((char *)&Opt->Audio.DeviceNameRec,299);
      continue;
    }

    /* FULL DUPLEX */
    if((ptmp=strstr((char *)&tmp,"FULL_DUPLEX="))!=NULL)
    {
      ptmp+=12;
      Items++;
      if(strstr(ptmp,"YES")!=NULL)
        Opt->Audio.FullDuplex=YES;
      else
        Opt->Audio.FullDuplex=NO;
      continue;
    }

    /* BITS */
    if((ptmp=strstr((char *)&tmp,"BITS="))!=NULL)
    {
      ptmp+=5;
      Items++;
      if(strstr(ptmp,"8")!=NULL)
        Opt->Audio.Bits=8;
      else
        Opt->Audio.Bits=16;
      continue;
    }

    /* SAMPLE RATE */
    if((ptmp=strstr((char *)&tmp,"SAMPLE_RATE="))!=NULL)
    {
      ptmp+=12;
      Items++;
      if(strstr(ptmp,"8000")!=NULL)
        Opt->Audio.SampleRate=8000;
      else
      {
        if(strstr(ptmp,"11025")!=NULL)
          Opt->Audio.SampleRate=11025;
         else
           if(strstr(ptmp,"22050")!=NULL)
             Opt->Audio.SampleRate=22050;
           else
             Opt->Audio.SampleRate=44100;
      }
      continue;
    }

    /* CHANNELS */
    if((ptmp=strstr((char *)&tmp,"CHANNELS="))!=NULL)
    {
      ptmp+=9;
      Items++;
      if(strstr(ptmp,"1")!=NULL)
        Opt->Audio.Channels=1;
      else
        Opt->Audio.Channels=2;
      continue;
    }

    /* REC MODE */
    if((ptmp=strstr((char *)&tmp,"RECMODE="))!=NULL)
    {
      ptmp+=8;
      Items++;
      if(strstr(ptmp,"direct")!=NULL)
        Opt->Audio.RecMode=RECDIRECT;
      else
        Opt->Audio.RecMode=RECMEMORY;
      continue;
    }

    /* PLAY MODE */
    if((ptmp=strstr((char *)&tmp,"PLAYMODE="))!=NULL)
    {
      ptmp+=9;
      Items++;
      if(strstr(ptmp,"direct")!=NULL)
        Opt->Audio.PlayMode=PLAYDIRECT;
      else
        if(strstr(ptmp,"mixfile")!=NULL)
          Opt->Audio.PlayMode=PLAYMIXFILE;
        else
          Opt->Audio.PlayMode=PLAYMEMORY;
      continue;
    }

  } // END while not feof()

  if(Items!=CFG_FILE_ITEMS)
    return -1;
  else
    return 0;
}













/*****************************************************
 *** CreateConfigFile
 *** creates a ~/.XhdREC/XhdREC.cfg file with
 *** default values
 *****************************************************/
int CreateConfigFile()
{
  char tmp[300];
  FILE *CfgFile;
  
  sprintf((char *)&tmp,"%s/.XhdREC.cfg",Opt->Prg.HomeDir);
  
  if( (CfgFile=fopen((char *)&tmp,"w"))==NULL )
  {
    StoreErrorMsg("can not create config file",ERR_WARNING);
    return -1;
  }

  _WriteSettingsToFile(CfgFile);

  fclose(CfgFile);
  return 0;
}




void _WriteSettingsToFile(FILE *CfgFile)
{
  char tmp[20];
  fprintf(CfgFile,"# do not edit, except XhdREC\n");
  fprintf(CfgFile,"# told you to do so ! ;-)\n\n");
  fprintf(CfgFile,"###################\n");
  fprintf(CfgFile,"# PROGRAM OPTIONS #\n");
  fprintf(CfgFile,"###################\n");
  CutLF(Opt->Prg.AudioDir,299);
  fprintf(CfgFile,"AUDIO_DIR=%s\n",Opt->Prg.AudioDir);
  CutLF(Opt->Prg.MixerApp,299);
  fprintf(CfgFile,"MIXER_APP=%s\n",Opt->Prg.MixerApp);
  fprintf(CfgFile,"HIDE_WINS_WHILE_REC=%s\n",(Opt->Prg.HideWins == YES) ? "YES" : "NO");
  fprintf(CfgFile,"REC_COUNTDOWN=%d\n",Opt->Prg.RecCountdown);
  fprintf(CfgFile,"###################\n");
  fprintf(CfgFile,"#  AUDIO OPTIONS  #\n");
  fprintf(CfgFile,"###################\n");
  CutLF(Opt->Audio.DeviceNamePlay,299);
  fprintf(CfgFile,"PLAY_DEV=%s\n",Opt->Audio.DeviceNamePlay);
  CutLF(Opt->Audio.DeviceNameRec,299);
  fprintf(CfgFile,"REC_DEV=%s\n",Opt->Audio.DeviceNameRec);
  fprintf(CfgFile,"FULL_DUPLEX=%s\n",(Opt->Audio.FullDuplex==YES) ? "YES" : "NO");
  fprintf(CfgFile,"BITS=%d\n",Opt->Audio.Bits);
  fprintf(CfgFile,"SAMPLE_RATE=%ld\n",Opt->Audio.SampleRate);
  fprintf(CfgFile,"CHANNELS=%d\n",Opt->Audio.Channels);
  fprintf(CfgFile,"RECMODE=%s\n",(Opt->Audio.RecMode == RECDIRECT) ? "direct" : "memory");
  if(Opt->Audio.PlayMode == PLAYDIRECT)
    sprintf((char *)&tmp,"direct");
  else
    if(Opt->Audio.PlayMode == PLAYMIXFILE)
      sprintf((char *)&tmp,"mixfile");
    else
      sprintf((char *)&tmp,"memory");
  fprintf(CfgFile,"PLAYMODE=%s\n",(char *)&tmp);
  fprintf(CfgFile,"###################\n");
  fprintf(CfgFile,"#      OTHER      #\n");
  fprintf(CfgFile,"###################\n");
  fprintf(CfgFile,"DISABLED -> VU_METER_STARTUP\n");
  fprintf(CfgFile,"DISABLED -> TRACK_WIN_STARTUP\n");
}






FILE* OpenFileForRec(char *FileName)
{
  return(fopen(FileName,"w"));
}

FILE* OpenFileForPlay(char *FileName)
{
  return(fopen(FileName,"r"));
}

void CloseFile(FILE *File)
{
  fclose(File);
}

void PanicCloseAllFiles()
{
  struct tAudioTrack *tmp1=TrackList;

  while(tmp1!=NULL)
  {
    if(tmp1->File != NULL)
    {
      CloseFile(tmp1->File);
      tmp1->File=NULL;
    }
    tmp1=tmp1->next;
  }
}

int OpenAllFiles(int Mode)
{
  struct tAudioTrack *tmp1=TrackList;
  char Msg[300];
  int err=0;

  while(tmp1!=NULL)
  {
    if((tmp1->Action==PLAY)&&( (Mode & PLAY)==PLAY ) )
    {
      if((tmp1->File=OpenFileForPlay((char *)&tmp1->FileName))==NULL)
      {
        err=1;
        break;
      }
#ifdef DEBUG
      else
        fprintf(stderr,"PLAY: %s\n",(char *)&tmp1->FileName);      
#endif
    }

    if((tmp1->Action==REC)&&( (Mode & REC) == REC ) )
    {
      if((tmp1->File=OpenFileForRec((char *)&tmp1->FileName))==NULL)
      {
        err=1;
        break;
      }
#ifdef DEBUG
      else
        fprintf(stderr,"REC : %s\n",(char *)&tmp1->FileName);      
#endif
    }
    tmp1=tmp1->next;
  }

  if(err!=0)
  {
    PanicCloseAllFiles();
    sprintf((char *)&Msg,"can't open %s",tmp1->FileName);
    StoreErrorMsg((char *)&Msg,ERR_CRITICAL);
    return -1;
  }

  return 0;
}


int GetDataFromFileToLong(FILE *File, long *Buffer,long MaxSize)
{
  long i;
  short j;

  
  if(File==NULL)
    return 0;

  if(feof(File))
    return -1;

  if(Opt->Audio.Bits==BIT16)
  {
    for(i=0;i<MaxSize;i+=2)
    {
      if(!feof(File))
      {
        j=((fgetc(File)) | ((fgetc(File) & 0xff ) << 8));
        *(Buffer++)=j*0.3;
      }
      else
        (short)*(Buffer++)=0;
    }
  }
  else  // 8 BIT
  {
    for(i=0;i<MaxSize;i++)
    {
      if(!feof(File))
      {
        j=fgetc(File); // | ((fgetc(File) & 0xff ) << 8));
        *(Buffer++)=(j-128)*0.3;
      }
      else
        (short)*(Buffer++)=0;
    }
  }

  if(feof(File))
    return -1;
  else
    return 0;
}


int AddDataFromFileToLong(FILE *File, long *Buffer,long MaxSize)
{
  long i;
  short j;

  if(File==NULL)
    return 0;

  if(feof(File))
    return -1;

  if(Opt->Audio.Bits==BIT16)
  {
    for(i=0;i<MaxSize;i+=2)
    {
      if(!feof(File))
      {
        j=((fgetc(File)) | ((fgetc(File) & 0xff ) << 8));
        (short)*(Buffer++)+=(j*0.3);
      }
    }
  }
  else  // 8 BIT
  {
    for(i=0;i<MaxSize;i++)
    {
      if(!feof(File))
      {
        j=fgetc(File); // | ((fgetc(File) & 0xff ) << 8));
        (short)*(Buffer++)+=(j-128)*0.3;
      }
    }
  }

  if(feof(File))
    return -1;
  else
    return 0;
  
}


int WriteBufferToFile(FILE *File, unsigned char *Buffer,long MaxSize)
{
  size_t i=sizeof(unsigned char);
  return(fwrite((unsigned char *)Buffer,i,MaxSize,File));
}




int ReadProjectFile(char *Name)
{
  char tmp1[300],tmp2[300],tmp3[300],*ptmp1,*ptmp2;
  FILE *PrjFile,*TrackDummy;
  struct tAudioTrack *tmp;
  int i;
  int PercentPerTrack;
  int MaxTracks=0;
  
  
  if( (PrjFile=fopen(Name,"r"))==NULL )
  {
    StoreErrorMsg("can not load project file",ERR_WARNING);
    return -1;
  }


  if(_ReadSettingsOutOfFile(PrjFile)!=0)
  {
    StoreErrorMsg("project settings invalid",ERR_WARNING);
    fclose(PrjFile);
    return -1;
  }

  UpdateProgramSettingsWin();
  UpdateAudioSettingsWin();
  RepaintWaitWin();


  sprintf((char *)&tmp1,"TRACK_");

  rewind(PrjFile);

  while(!feof(PrjFile))
  {
    fgets((char *)&tmp2,299,PrjFile);
    if(strstr((char *)&tmp2,(char *)&tmp1)!=NULL)
      MaxTracks++;    
  }
  
  if(MaxTracks!=0)
    PercentPerTrack=100/MaxTracks;
  else
    PercentPerTrack=0;

#ifdef DEBUG
  fprintf(stderr,"TRACKS: %d\n",MaxTracks);
#endif

  /* now scan for tracks in the right order */
  for(i=1;i<=MAXFILES;i++)
  {
    rewind(PrjFile);

    while(!feof(PrjFile))
    {
      sprintf((char *)&tmp1,"TRACK_%02d=",i);
      fgets((char *)&tmp2,299,PrjFile);

      /* TODO */
      fl_check_forms();

      if((ptmp1=strstr((char *)&tmp2,(char *)&tmp1))!=NULL)
      {
        ptmp1+=9;
        CutLF(ptmp1,289);
        if((TrackDummy=OpenFileForPlay(ptmp1))!=NULL)
        {
          CloseFile(TrackDummy);
          sprintf((char *)&tmp1,"NAME_%02d=",i);
          fgets((char *)&tmp3,299,PrjFile);
          if((ptmp2=strstr((char *)&tmp3,(char *)&tmp1))!=NULL)
          {
            ptmp2+=8;
            CutLF(ptmp2,289);        
          }
          else
          {
            sprintf((char *)&tmp3,"MY TRACK");
            ptmp2=(char *)&tmp3;
          }
          tmp=CreateTrack(ptmp1,ptmp2);
          if(tmp!=NULL)
          {
            /* TODO */
            SetWaitWinProgress((Opt->Prg.WaitWinProgress+=PercentPerTrack));

            AddTrackToList(tmp);
            CreateTrackPeakFile(tmp,PEAK_SAMPLE_SKIP);
            LoadPeakFileToMemory(tmp);
            RepaintTrackWin();
          }
        }  // END if NAME_%02d found
        else
        {
          sprintf((char *)&tmp1,"can not find file:\n%s",ptmp1);
          fl_show_messages((char *)&tmp1);
        }  // ENDE else NAME_%02 not found in next line
        break;
      }  // END if TRACK_%d= found
    }  // END while !end of file
  }  // END for maximum number of tracks


  fclose(PrjFile);

  /* store the project filename */
  sprintf((char *)&Opt->Prg.ProjectFileName,"%s",Name);
  CutLastSlash(Name);
  /* this is our new working dir */
  sprintf((char *)&Opt->Prg.AudioDir,"%s",Name);

  return 0;
}





int CreateProjectFile(char *Name)
{
  FILE *PrjFile;
  struct tAudioTrack *tmp=TrackList;
  int i;
  
  
  if( (PrjFile=fopen(Name,"w"))==NULL )
  {
    StoreErrorMsg("can not save project file",ERR_WARNING);
    return -1;
  }

  _WriteSettingsToFile(PrjFile);

  fprintf(PrjFile,"###################\n");
  fprintf(PrjFile,"#     TRACKS      #\n");
  fprintf(PrjFile,"###################\n");
 
  i=1;
  while(tmp!=NULL)
  {
    fprintf(PrjFile,"TRACK_%02d=%s\n",i,tmp->FileName);
    fprintf(PrjFile,"NAME_%02d=%s\n",i++,tmp->Name);
    tmp=tmp->next;
  }

  fclose(PrjFile);
  return 0;
}



off_t GetFileSize(char *FileName)
{
  struct stat FileStats;

  if(stat(FileName,&FileStats)!=0)
    return 0;

  return FileStats.st_size;
}



int CreateTrackPeakFile(struct tAudioTrack *Track, long SampleSkip)
{
  FILE *in,*out;
  char tmp[300];
  short VolL,VolR,Help;
  long i;

  if((in=OpenFileForPlay(Track->FileName))==0)
  {
    StoreErrorMsg("unable to open track for creating peak file",ERR_CRITICAL);
    return -1;
  }

  sprintf((char *)&tmp,"%s",Track->FileName);
  AddExtensionString((char *)&tmp,".pk");
  if((out=OpenFileForRec((char *)&tmp))==0)
  {
    CloseFile(in);
    StoreErrorMsg("unable to create peak file",ERR_CRITICAL);
    return -1;
  }



  while(!feof(in))
  {
    VolL=VolR=0;
    for(i=0;i<SampleSkip;i++)
    {
      /* read the first channel */
      if(Opt->Audio.Bits==BIT8)
        VolL=fgetc(in)-128;
      else
      {
        Help=((fgetc(in)) | ((fgetc(in) & 0xff ) << 8));
        Help/=257;
        if( abs(Help) > abs(VolL))
          VolL=Help;
      }
      /* read the second channel, if necessary */
      if(Opt->Audio.Channels==2)
      {
        if(Opt->Audio.Bits==BIT8)
          VolR=fgetc(in)-128;
        else
        {
          Help=((fgetc(in)) | ((fgetc(in) & 0xff ) << 8));
          Help/=257;
          if( abs(Help) > abs(VolR))
            VolR=Help;
        }
      }
      
    }  // END for scan highest value

    fputc((char) VolL,out);
    if(Opt->Audio.Channels==2)
      fputc((char) VolR,out);
    
    if(feof(in))
      break;
  }

  CloseFile(in);
  CloseFile(out);
  return 0;
}



FILE *OpenPeakFile(struct tAudioTrack *Track)
{
  char tmp[300];
  sprintf((char *)&tmp,"%s",Track->FileName);
  AddExtensionString((char *)&tmp,".pk");
  return OpenFileForPlay((char *)&tmp);
}


int LoadPeakFileToMemory(struct tAudioTrack *Track)
{
  FILE *in;
  char tmp[300],*tmp1;
  off_t FileSize;

  if(Track==NULL)
    return 0;

  sprintf((char *)&tmp,"%s",Track->FileName);
  AddExtensionString((char *)&tmp,".pk");


  if((FileSize=GetFileSize((char *)&tmp))==0)
  {
    StoreErrorMsg("unable to read peak file size",ERR_CRITICAL);
    return -1;
  }


  if((in=OpenFileForPlay((char *)&tmp))==0)
  {
    StoreErrorMsg("unable to open peak file",ERR_CRITICAL);
    return -1;
  }


  if(Track->PeakSamples!=NULL)
  {
    free(Track->PeakSamples);
    Track->PeakSamples=NULL;
    Track->PeakSampleLenght=0;
  }


  if((tmp1=Track->PeakSamples=(char *)malloc(10 + FileSize * sizeof(char)))==NULL)
  {
    fclose(in);
    StoreErrorMsg("unable to create peak file buffer",ERR_CRITICAL);
    return -1;
  }


  Track->PeakSampleLenght=FileSize;

  while(!feof(in))
    *(tmp1++)=(char)fgetc(in);

  fclose(in);
  return 0;
}








