#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "forms.h"
#include "xwin2.h"
#include "global.h"
#include "audio.h"
#include "error.h"
#include "tools.h"
#include "file.h"
#include "windows.h"


void _DeleteTrack(struct tAudioTrack *ThisOne);



struct tAudioTrack* CreateTrack(char *FileName, char *Name)
{
  struct tAudioTrack *NewTrack;

  NewTrack=(struct tAudioTrack *) malloc(sizeof(struct tAudioTrack));
  if(NewTrack!=NULL)
  {
    CutLF(Name,299);
    sprintf((char *)&NewTrack->Name,"%s",Name);
    sprintf((char *)&NewTrack->FileName,"%s",FileName);
    NewTrack->File=NULL;
    NewTrack->Action=MUTE;
    NewTrack->Volume=0;
    NewTrack->Pan=0;
    NewTrack->PeakSamples=NULL;
    NewTrack->PeakSampleLenght=0;
    NewTrack->Selected=NO;
    NewTrack->next=NULL;
  }
  return NewTrack;
}


void AddTrackToList(struct tAudioTrack *ThisOne)
{
  struct tAudioTrack *tmp=NULL;

  if(ThisOne==NULL)
    return;

  if(TrackList==NULL)
  {
    TrackList=ThisOne;
    Opt->Tracks.NrOfTracks++;
    return;
  }

  tmp=TrackList;
  while(tmp->next!=NULL)
    tmp=tmp->next;
  /* now, tmp points on the last element */
  tmp->next=ThisOne;
  Opt->Tracks.NrOfTracks++;
}


void RemoveTrackFromList(struct tAudioTrack *ThisOne)
{
  struct tAudioTrack *tmp=TrackList;

  if((ThisOne==NULL)||(TrackList==NULL))
    return;

  if(tmp==ThisOne)       /* it's the first element */
  {
    TrackList=ThisOne->next;   /* could also be NULL (no further element) */
  }
  else                   /* all other elements */
  {
    while((tmp->next!=ThisOne)&&(tmp!=NULL))
      tmp=tmp->next;
    if(tmp==NULL)        /* element is not in list */
      return;
    tmp->next=tmp->next->next; /* could also be NULL (no further element) */
  }

  _DeleteTrack(ThisOne);
  Opt->Tracks.NrOfTracks--;
}





void RemoveAllTracksFromList()
{
  struct tAudioTrack *tmp1,*tmp2;
  tmp1=TrackList;
  
  while(tmp1!=NULL)
  {
    tmp2=tmp1->next;
    RemoveTrackFromList(tmp1);
    tmp1=tmp2;
  }
  TrackList=NULL;
  Opt->Tracks.NrOfTracks=0;
}


void _DeleteTrack(struct tAudioTrack *ThisOne)
{
  /* dangerous function!
   * must only be used with UNCHAINED elements !! */
  if(ThisOne==NULL)
    return;

  if(ThisOne->PeakSamples!=0)
    free((char *)ThisOne->PeakSamples);

  free(ThisOne);
}



int FileAlreadyInProject(char *FileName)
{
  struct tAudioTrack *tmp1=TrackList;
  
  while(tmp1!=NULL)
  {
    if(strcmp(FileName,(char *)&tmp1->FileName)==0)
      return 1;
    tmp1=tmp1->next;
  }
  return 0;
}



int GetNecessaryPlayMode()
{
  int Mask=0;
  struct tAudioTrack *tmp1=TrackList;

  while(tmp1!=NULL)
  {
    if(tmp1->Action==PLAY)
      Mask=Mask | PLAY;
    if(tmp1->Action==REC)
      Mask=Mask | REC;
    tmp1=tmp1->next;
  }
  return Mask;
}









/***************************************************************************
 *** PlayMultipleTracks
 *** 
 *** 
 ***************************************************************************/
int PlayMultipleTracks2()
{
  long *Cache,*Cptr,CacheSize;
  unsigned char *Buffer,*Bptr;
  struct tAudioTrack *tmp1=TrackList;
  long i,j;        /* long in case of large cache! */
  FILE *FileList[MAXFILES];
  int FilesOpen,FilesMax;

  /* TESTING */
  /* store global data in local vars so that
   * the compiler can optimize them (register)
   * (not really needed, but every nano-second counts ;-) */
  int FragSize=Opt->Audio.PlayFragsize;
  int Fragments=Opt->Audio.PlayFragments;
  int FirstRead=0;


  /* open all files */
  if(OpenAllFiles(PLAY)!=0)
  {
    PanicCloseAllFiles();
    return -1;
  }

  i=0;

  /* assign open files (PLAYMODE) to local FileList */
  while(tmp1!=NULL)
  {
    if(tmp1->Action==PLAY)
      FileList[i++]=tmp1->File;
    tmp1=tmp1->next;
  }

  /* rest of list -> NULL */
  FilesOpen=FilesMax=i;
  for(j=i;j<MAXFILES;j++)
    FileList[j]=NULL;
  

#ifdef DEBUG
  fprintf(stderr,"FILES OPENED\n");
#endif

  /* create a disk-read cache 
   * it's size is the complete count of bytes that can be wriiten to the sounddev
   * without blocking it (FragSize * Fragments), multiplied by a const factor.
   * Now, we can read more data in the same time (harddisk/memory-cache !).
   * Notice that this cache's type is <long> (and not unsigned char)!
   * It's used for mixing too ! Therefore the normal size is divided by 2
   * (I simply left out the factor <Fragments> ) */
  Cache=Cptr=(long *)malloc((CacheSize=Fragments * FragSize * sizeof(long)));
  if(Cache==NULL)
  {
    StoreErrorMsg("can't allocate cache size !",ERR_CRITICAL);
    PanicCloseAllFiles();
    return -1;
  }
#ifdef DEBUG
  else
    fprintf(stderr,"DYNAMIC CACHE INITIALIZED: %lu !\n",CacheSize);
#endif  

  /* create an output buffer
   * it's size is the size of Cache multiplied by 2, because now we're using
   * unsigned chars ! */
  Buffer=Bptr=(unsigned char *)malloc(2 * Fragments * FragSize * sizeof(unsigned char));
  if(Buffer==NULL)
  {
    StoreErrorMsg("can't allocate buffer size !",ERR_CRITICAL);
    free(Cache);
    PanicCloseAllFiles();
    return -1;
  }
#ifdef DEBUG
  else
    fprintf(stderr,"DYNAMIC BUFFER INITIALIZED!\n");
#endif  


  /* main play loop:
   * read data from disk, mix it and play through sounddev */

  while(FilesOpen!=0)      /* until no more data to read */
  {
    /* clear cache */
    /* COULD BE BETTER !
     * 2nd function instead of AddDataFromFileToLong:
     * should not ADD but OVERWRITE contents of cache ! */
    Cptr=Cache;
    for(i=0;i<FragSize;i++)
      *(Cptr++)=0;

    /* read all files and mix them */
    FirstRead=1;
    for(i=0;i<FilesMax;i++)
    {
      if((FirstRead--)==1)
      {
        if(GetDataFromFileToLong(FileList[i],Cache,FragSize*Fragments)==-1)
        {
          FileList[i]=NULL;
          FilesOpen--;
        }
      }
      else
        if(AddDataFromFileToLong(FileList[i],Cache,FragSize*Fragments)==-1)
        {
          FileList[i]=NULL;
          FilesOpen--;
        }
    }

    /* now (re)create a little endian format for the sounddev driver and play it */
    Cptr=Cache; Bptr=Buffer;
    if(Opt->Audio.Bits==BIT16)
    {
      for(i=0;i<FragSize*Fragments;i+=2)
      {    
        *(Bptr++)=~(unsigned char)(*Cptr & 0xff);              /* low byte  */
        *(Bptr++)=~(unsigned char)((*(Cptr++) >> 8) & 0xff);   /* high byte */
      }
    }
    else
    {
      for(i=0;i<FragSize*Fragments;i++)
      {    
        *(Bptr++)=(~(unsigned char)(*(Cptr++) & 0xff))+128;      /* low byte  */

//        *(Bptr++)=~(unsigned char)((*(Cptr++) >> 8) & 0xff);   /* high byte */
      }
    }

    for(i=0;i<Fragments;i++)
    {
      PlayBufferToDevice(Buffer+i*FragSize,FragSize);
    }

    /* TESTING */
    fl_check_forms();
    if(Opt->Prg.StopAction==YES)
      break;

  }  // END while still something to play

  /* wait until buffer is played, otherwise we'll get an segfault */
  FlushSoundDevice();


  /* clear Cache, Buffer and close all previously opened files */
  free((long *)Cache);  
  free((unsigned char *)Buffer);
  PanicCloseAllFiles();
  return 0;
}






/***************************************************************************
 *** RecMultipleTracks
 *** 
 *** 
 ***************************************************************************/
int RecMultipleTracks2()
{
  unsigned char *Buffer,*Bptr;
  struct tAudioTrack *tmp1=TrackList;

  /* TODO */
  struct tAudioTrack *RecTrack;
  
  long i,j;        /* long in case of large cache! */
  FILE *FileList[MAXFILES];              /* only 1 used (future exp.) */
  int FilesOpen,FilesMax;
  int FragSize=Opt->Audio.RecFragsize;
  int Fragments=Opt->Audio.RecFragments;

  /* open all files */
  if(OpenAllFiles(REC)!=0)
  {
    PanicCloseAllFiles();
    return -1;
  }

  i=0;

  /* assign open files (RECMODE) to local FileList */
  while(tmp1!=NULL)
  {
    if(tmp1->Action==REC)
    {
      FileList[i++]=tmp1->File;
      /* TODO */
      RecTrack=tmp1;
    }
    tmp1=tmp1->next;
  }

  /* rest of list -> NULL */
  FilesOpen=FilesMax=i;
  for(j=i;j<MAXFILES;j++)
    FileList[j]=NULL;
  

  /* a hack for future expansions */
  if(FilesOpen>1)
  {
    PanicCloseAllFiles();
    StoreErrorMsg("you can't record more than 1 track\nat the same time",ERR_CRITICAL);
    return -1;
  }

#ifdef DEBUG
  fprintf(stderr,"FILES OPENED\n");
#endif

  Buffer=Bptr=(unsigned char *)malloc(Fragments * FragSize * sizeof(unsigned char));
  if(Buffer==NULL)
  {
    StoreErrorMsg("can't allocate buffer size !",ERR_CRITICAL);
    PanicCloseAllFiles();
    return -1;
  }
#ifdef DEBUG
  else
    fprintf(stderr,"DYNAMIC BUFFER INITIALIZED!\n");
#endif  

  /* main rec loop */
  for(;;)
  {
    for(i=0;i<Fragments;i++)
      RecBufferFromDevice(Buffer+i*FragSize,FragSize); 

    WriteBufferToFile(FileList[0],Buffer,FragSize*Fragments);   

    /* TESTING */
    fl_check_forms();
    if(Opt->Prg.StopAction==YES)
      break;
  }

  /* clear Cache, Buffer and close all previously opened files */
  free((unsigned char *)Buffer);
  PanicCloseAllFiles();

  /* TODO
   * this needs to be strored somewhere else */
  CreateTrackPeakFile(RecTrack,PEAK_SAMPLE_SKIP);
  LoadPeakFileToMemory(RecTrack);

  
  return 0;

}

/***************************************************************************
 *** PlayAndRecMultipleTracks
 *** 
 *** 
 ***************************************************************************/
int PlayAndRecMultipleTracks2()
{
  long *PlayCache,*PCptr,PlayCacheSize;
  unsigned char *PlayBuffer,*PBptr;
  unsigned char *RecBuffer,*RBptr;
  struct tAudioTrack *tmp1=TrackList;

  /* TODO */
  struct tAudioTrack *RecTrack;

  long i,j;  
  FILE *PlayFileList[MAXFILES];
  FILE *RecFileList[MAXFILES];       /* only 1 used (future exp.) */
  int PlayFilesOpen,PlayFilesMax;
  int FragSize=Opt->Audio.DuplexPlayFragsize;
  int Fragments=Opt->Audio.DuplexPlayFragments;
  int Factor=Fragments*FragSize;
  int FirstRead=0;
  char ErrMsg[40];
  int FragmentWarning=NO;


  /* First check if the device's play and rec buffer have equal size. */
/*
  if(FragSize!=Opt->Audio.RecFragsize)
  {
    sprintf((char *)&ErrMsg,"buffer sizes not equal:\nplaybuf: %d\n recbuf: %d\ntry another full-duplex buffer size",
            FragSize,Opt->Audio.RecFragsize);
    StoreErrorMsg((char *)&ErrMsg,ERR_CRITICAL);
    return -1;
  }
*/

  /* set Fragements to the lowest Fragsize of play and rec fragments */
  /* looks strange, but:
   * if Fragments (the PlayFragments) is not equal to RecFragments, we only have
   * to reinitialize it, if it's bigger than RecFragments. Otherwise it already
   * contains the lowest value. */
/*
  if(Fragments!=Opt->Audio.RecFragments)
  {
    Fragments=Opt->Audio.RecFragments;
    FragmentWarning=YES;
  }
*/

  /* open all files */
  if(OpenAllFiles(PLAY | REC)!=0)
  {
    PanicCloseAllFiles();
    return -1;
  }


  i=0; j=0;

  /* assign open files (PLAYMODE) to local FileList */
  while(tmp1!=NULL)
  {
    if(tmp1->Action==PLAY)
      PlayFileList[i++]=tmp1->File;
    if(tmp1->Action==REC)
    {
      RecFileList[j++]=tmp1->File;
      /* TODO */
      RecTrack=tmp1;
    }
    tmp1=tmp1->next;
  }
  PlayFilesOpen=PlayFilesMax=i;

  /* a hack for future expansions */
  if(j>1)
  {
    PanicCloseAllFiles();
    StoreErrorMsg("you can't record more than 1 track\nat the same time",ERR_CRITICAL);
    return -1;
  }

  /* rest of list -> NULL */
  for(i=PlayFilesOpen;i<MAXFILES;i++)
    PlayFileList[i]=NULL;
  for(i=j;i<MAXFILES;i++)
    RecFileList[i]=NULL;
  
#ifdef DEBUG
  fprintf(stderr,"FILES OPENED: %d\n",PlayFilesOpen+(int)j);
#endif
  PlayCache=PCptr=(long *)malloc((PlayCacheSize=Factor * sizeof(long)));
  if(PlayCache==NULL)
  {
    StoreErrorMsg("can't allocate cache size !",ERR_CRITICAL);
    PanicCloseAllFiles();
    return -1;
  }
#ifdef DEBUG
  else
    fprintf(stderr,"DYNAMIC CACHE INITIALIZED: %lu !\n",PlayCacheSize);
#endif  
  PlayBuffer=PBptr=(unsigned char *)malloc(2 * Factor * sizeof(unsigned char));
  if(PlayBuffer==NULL)
  {
    StoreErrorMsg("can't allocate play buffer size !",ERR_CRITICAL);
    free(PlayCache);
    PanicCloseAllFiles();
    return -1;
  }
#ifdef DEBUG
  else
    fprintf(stderr,"DYNAMIC PLAY BUFFER INITIALIZED!\n");
#endif  


  RecBuffer=RBptr=(unsigned char *)malloc(2 * Factor * sizeof(unsigned char));
  if(RecBuffer==NULL)
  {
    StoreErrorMsg("can't allocate play buffer size !",ERR_CRITICAL);
    free(PlayCache);
    free(PlayBuffer);
    PanicCloseAllFiles();
    return -1;
  }
#ifdef DEBUG
  else
    fprintf(stderr,"DYNAMIC PLAY BUFFER INITIALIZED!\n");
#endif  


  /* main play/rec loop:
   * read data from disk, mix it and play through sounddev */
  for(;;)
  {
    /* read all play-files and mix them */
    FirstRead=1;
    for(i=0;i<PlayFilesMax;i++)
    {
      if((FirstRead--)==1)
      {
        if(GetDataFromFileToLong(PlayFileList[i],PlayCache,Factor)==-1)
        {
          PlayFileList[i]=NULL;
          PlayFilesOpen--;
        }
      }
      else
        if(AddDataFromFileToLong(PlayFileList[i],PlayCache,Factor)==-1)
        {
          PlayFileList[i]=NULL;
          PlayFilesOpen--;
        }
    } // END for any playfile left...

    /* now (re)create a little endian format for the sounddev driver and play it */
    PCptr=PlayCache; PBptr=PlayBuffer;
    if(Opt->Audio.Bits==BIT16)
    {
      for(i=0;i<Factor;i+=2)
      {    
        *(PBptr++)=(unsigned char)(*PCptr & 0xff);              /* low byte  */
        *(PBptr++)=(unsigned char)((*(PCptr++) >> 8) & 0xff);   /* high byte */
      }
    }
    else
    {
      for(i=0;i<Factor;i++)
      {    
        *(PBptr++)=((unsigned char)(*(PCptr++) & 0xff))+128;              /* low byte  */
//        *(PBptr++)=(unsigned char)((*(PCptr++) >> 8) & 0xff);   /* high byte */
      }
    }

    for(i=0;i<Fragments;i++)
    {
      PlayBufferToDevice(PlayBuffer+i*FragSize,FragSize);
      RecBufferFromDevice(RecBuffer+i*FragSize,FragSize); 
    }
    WriteBufferToFile(RecFileList[0],RecBuffer,Factor);   

    /* TESTING */
    fl_check_forms();
    if(Opt->Prg.StopAction==YES)
      break;


  }  // END for main play/rec loop

  FlushSoundDevice();
  free((long *)PlayCache);  
  free((unsigned char *)PlayBuffer);
  free((unsigned char *)RecBuffer);
  PanicCloseAllFiles();

  /* TODO
   * this needs to be strored somewhere else */
  CreateTrackPeakFile(RecTrack,PEAK_SAMPLE_SKIP);
  LoadPeakFileToMemory(RecTrack);


  if(FragmentWarning==YES)
  {
    StoreErrorMsg("play and rec fragments were not equal during recording\n(performance reduced)",ERR_WARNING);
    return -1;
  }

  return 0;
}





void RemoveTrackFromProject()
{
  struct tAudioTrack *tmp=TrackList;

  while(tmp!=NULL)
  {
    if(tmp->Selected==YES)
    {
      RemoveTrackFromList(tmp);
      tmp=TrackList;
      continue;
    }
    tmp=tmp->next;
  }
  RepaintTrackWin();
}




void AddTrackToProject()
{
  struct tAudioTrack *tmp;
  char *FileName,*Name;
  FILE *File;

  if(Opt->Tracks.NrOfTracks>=MAXFILES)
  {
    fl_show_messages("maximum number of tracks reached");
    return;
  }
  FileName=(char *)fl_show_fselector("",(char *)&Opt->Prg.AudioDir,"*.xhd","");
  if(FileName!=0)
  {
    AddExtensionString(FileName,".xhd");
  
    if(!FileAlreadyInProject(FileName))
    {
      if((File=OpenFileForPlay(FileName))!=NULL)
      {
        CloseFile(File);
        if(fl_show_question("file exists! overwrite ???",2)==0)
          return;
      }

      if((File=OpenFileForRec(FileName))!=NULL)
      {
        CloseFile(File);
        Name=(char *)fl_show_input("enter a name for the track:","MY TRACK");
        CutLF(Name,15);
        tmp=CreateTrack(FileName,Name);
        if(tmp!=NULL)
        {
          AddTrackToList(tmp);
          CreateTrackPeakFile(tmp,PEAK_SAMPLE_SKIP);
          LoadPeakFileToMemory(tmp);
        }
        RepaintTrackWin();
      }
      else
        fl_show_messages("can not create file");
    }
    else
      fl_show_messages("this file is already used by your project!");
  }
}






void ImportTrackToProject()
{
  struct tAudioTrack *tmp;
  char *FileName,*Name;
  FILE *File;

  if(Opt->Tracks.NrOfTracks>=MAXFILES)
  {
    fl_show_messages("maximum number of tracks reached");
    return;
  }
  FileName=(char *)fl_show_fselector("",(char *)&Opt->Prg.AudioDir,"*.xhd","");
  if(FileName!=0)
  {
    if(!FileAlreadyInProject(FileName))
    {
      if((File=OpenFileForPlay(FileName))!=NULL)
      {
        CloseFile(File);
        Name=(char *)fl_show_input("enter a name for the track:","MY TRACK");
        CutLF(Name,15);
        tmp=CreateTrack(FileName,Name);
        if(tmp!=NULL)
        {
          AddTrackToList(tmp);
          CreateTrackPeakFile(tmp,PEAK_SAMPLE_SKIP);
          LoadPeakFileToMemory(tmp);
        }
        RepaintTrackWin();
      }
      else
        fl_show_messages("can not open file");
    }
    else
      fl_show_messages("this file is already used by your project!");
  }
}




long GetMaxPeakSampleLenght()
{
  struct tAudioTrack *tmp=TrackList;
  long MaxSize=0;

  while(tmp!=NULL)
  {
    if(tmp->PeakSampleLenght > MaxSize)
      MaxSize=tmp->PeakSampleLenght;
    tmp=tmp->next;
  }

  return MaxSize;
  
}









