#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "forms.h"
#include "xwin2.h"
#include "global.h"
#include "audio.h"
#include "error.h"
#include "file.h"
#include "tracks.h"




/* for track display */
#define XSKIP      150
#define XSKIPEND   10



int TrackWinExposeEvent(FL_OBJECT *Obj, Window win, int w, int h, XEvent *ev, void *d);
int TrackWinButtonEvent(FL_OBJECT *Obj, Window win, int w, int h, XEvent *ev, void *d);
int StopWinExposeEvent(FL_OBJECT *Obj, Window win, int w, int h, XEvent *ev, void *d);
int StopWinButtonEvent(FL_OBJECT *Obj, Window win, int w, int h, XEvent *ev, void *d);
int WaitWinExposeEvent(FL_OBJECT *Obj, Window win, int w, int h, XEvent *ev, void *d);
int WaitWinButtonEvent(FL_OBJECT *Obj, Window win, int w, int h, XEvent *ev, void *d);
int WaitWinVisible(void);


void RepaintTrackWin(void);
void RepaintWaitWin(void);
int IsInRectangle(int thisx, int thisy, int x1, int y1, int x2, int y2);






/*********************************************************
 *** UpdateProgramSettingsWin
 *** updates all settings in the Program-Settings Window
 *********************************************************/
void UpdateProgramSettingsWin()
{
  int i;

  fl_set_input(fd_fSetPrg->iPrgOptAudioDir,(const char *)&Opt->Prg.AudioDir);
  fl_set_input(fd_fSetPrg->iPrgOptMixerApp,(const char *)&Opt->Prg.MixerApp);
  fl_set_choice(fd_fSetPrg->chPrgOptHideWindows,(Opt->Prg.HideWins==YES) ? 1:2);
  if(Opt->Prg.RecCountdown==10) i=5;
  else
    if(Opt->Prg.RecCountdown==5) i=4;
    else
      if(Opt->Prg.RecCountdown==3) i=3;
      else
        if(Opt->Prg.RecCountdown==1) i=2;
        else i=1;
  fl_set_choice(fd_fSetPrg->chPrgOptRecCount,i);
}

/******************************************************************
 *** ReadProgramSettingsWin
 *** reads all settings of the Programs-Settings win and stores
 *** the values in the internal vars
 ******************************************************************/
void ReadProgramSettingsWin()
{
  int i;

  sprintf((char *)&Opt->Prg.AudioDir,"%s",fl_get_input(fd_fSetPrg->iPrgOptAudioDir));
  sprintf((char *)&Opt->Prg.MixerApp ,"%s",fl_get_input(fd_fSetPrg->iPrgOptMixerApp));
  Opt->Prg.HideWins=(fl_get_choice(fd_fSetPrg->chPrgOptHideWindows)==1) ? YES : NO;
  i=fl_get_choice(fd_fSetPrg->chPrgOptRecCount);
  if(i==1) Opt->Prg.RecCountdown=0;
  else
    if(i==2) Opt->Prg.RecCountdown=1;
    else
      if(i==3) Opt->Prg.RecCountdown=3;
      else
        if(i==4) Opt->Prg.RecCountdown=5;
        else Opt->Prg.RecCountdown=10;
}



/*******************************************************
 *** UpdateAudioSettingsWin
 *** updates all settings in the Audio-Settings Window
 *******************************************************/
void UpdateAudioSettingsWin()
{
  int i;
  char tmp[50];

  fl_set_input(fd_fSetAudio->iAudOptPlaybackDev,(const char *)&Opt->Audio.DeviceNamePlay);
  fl_set_input(fd_fSetAudio->iAudOptRecordDev,(const char *)&Opt->Audio.DeviceNameRec);
  fl_set_choice(fd_fSetAudio->chAudOptBits,(Opt->Audio.Bits==8) ? 1:2);
  fl_set_choice(fd_fSetAudio->chAudOptChannels,(Opt->Audio.Channels==1) ? 1:2);
  fl_set_choice(fd_fSetAudio->chAudOptFullDuplex,(Opt->Audio.FullDuplex==YES) ? 1:2);
  fl_set_choice(fd_fSetAudio->chAudOptRecMode,(Opt->Audio.RecMode==RECDIRECT) ? 1:2);
  if(Opt->Audio.PlayMode==PLAYDIRECT)
    i=1;
  else
    if(Opt->Audio.PlayMode==PLAYMIXFILE)
      i=2;
    else
      i=3;
  fl_set_choice(fd_fSetAudio->chAudOptPlayMode,i);
  if(Opt->Audio.SampleRate==44100) i=4;
  else
    if(Opt->Audio.SampleRate==22050) i=3;
    else
      if(Opt->Audio.SampleRate==11025) i=2;
      else i=1;
  fl_set_choice(fd_fSetAudio->chAudOptSampleRate,i);

  sprintf((char *)&tmp,"%d",Opt->Audio.PlayFragsize);
  fl_set_input(fd_fSetAudio->iAudOptPlayFragsize,(char *)&tmp);
  sprintf((char *)&tmp,"%d",Opt->Audio.PlayFragments);
  fl_set_input(fd_fSetAudio->iAudOptPlayFragments,(char *)&tmp);

  sprintf((char *)&tmp,"%d",Opt->Audio.RecFragsize);
  fl_set_input(fd_fSetAudio->iAudOptRecFragsize,(char *)&tmp);
  sprintf((char *)&tmp,"%d",Opt->Audio.RecFragments);
  fl_set_input(fd_fSetAudio->iAudOptRecFragments,(char *)&tmp);

  sprintf((char *)&tmp,"%d",Opt->Audio.DuplexPlayFragsize);
  fl_set_input(fd_fSetAudio->iAudOptDuplexPlayFragsize,(char *)&tmp);
  sprintf((char *)&tmp,"%d",Opt->Audio.DuplexPlayFragments);
  fl_set_input(fd_fSetAudio->iAudOptDuplexPlayFragments,(char *)&tmp);
  sprintf((char *)&tmp,"%d",Opt->Audio.DuplexRecFragsize);
  fl_set_input(fd_fSetAudio->iAudOptDuplexRecFragsize,(char *)&tmp);
  sprintf((char *)&tmp,"%d",Opt->Audio.DuplexRecFragments);
  fl_set_input(fd_fSetAudio->iAudOptDuplexRecFragments,(char *)&tmp);

}


/*****************************************************************
 *** ReadAudioSettingsWin
 *** reads all settings of the Audio Settings Windows and stores
 *** values in the internal vars
 *****************************************************************/
void ReadAudioSettingsWin()
{
  int i;

  sprintf((char *)&Opt->Audio.DeviceNamePlay,"%s",fl_get_input(fd_fSetAudio->iAudOptPlaybackDev));
  sprintf((char *)&Opt->Audio.DeviceNameRec,"%s",fl_get_input(fd_fSetAudio->iAudOptRecordDev));
  Opt->Audio.Bits=(fl_get_choice(fd_fSetAudio->chAudOptBits) == 1) ? 8 : 16;
  Opt->Audio.Channels=fl_get_choice(fd_fSetAudio->chAudOptChannels);
  Opt->Audio.FullDuplex=(fl_get_choice(fd_fSetAudio->chAudOptFullDuplex)==1) ? YES : NO;
  Opt->Audio.RecMode=(fl_get_choice(fd_fSetAudio->chAudOptRecMode)==1) ? RECDIRECT : RECMEMORY;
  i=fl_get_choice(fd_fSetAudio->chAudOptPlayMode);
  if(i==1)
    Opt->Audio.PlayMode=PLAYDIRECT;
  else
    if(i==2)
      Opt->Audio.PlayMode=PLAYMIXFILE;
    else
      Opt->Audio.PlayMode=PLAYMEMORY;
    
  i=fl_get_choice(fd_fSetAudio->chAudOptSampleRate);
  if(i==1) Opt->Audio.SampleRate=8000;
  else
    if(i==2) Opt->Audio.SampleRate=11025;
    else
      if(i==3) Opt->Audio.SampleRate=22050;
      else Opt->Audio.SampleRate=44100;

  Opt->Audio.PlayFragsize=atoi(fl_get_input(fd_fSetAudio->iAudOptPlayFragsize));
  Opt->Audio.PlayFragments=atoi(fl_get_input(fd_fSetAudio->iAudOptPlayFragments));
  Opt->Audio.RecFragsize=atoi(fl_get_input(fd_fSetAudio->iAudOptRecFragsize));
  Opt->Audio.RecFragments=atoi(fl_get_input(fd_fSetAudio->iAudOptRecFragments));
  Opt->Audio.DuplexPlayFragsize=atoi(fl_get_input(fd_fSetAudio->iAudOptDuplexPlayFragsize));
  Opt->Audio.DuplexPlayFragments=atoi(fl_get_input(fd_fSetAudio->iAudOptDuplexPlayFragments));
  Opt->Audio.DuplexRecFragsize=atoi(fl_get_input(fd_fSetAudio->iAudOptDuplexRecFragsize));
  Opt->Audio.DuplexRecFragments=atoi(fl_get_input(fd_fSetAudio->iAudOptDuplexRecFragments));
}




void ShowMenuWin()
{
  fl_show_form(fd_fMenu->fMenu,FL_PLACE_CENTERFREE,FL_FULLBORDER,"XhdREC");
}

void HideMenuWin()
{
  fl_hide_form(fd_fMenu->fMenu);
}









void ShowTrackWin()
{
  if(Opt->Prg.TrackWinVisible==YES)
    return;
  
  Opt->Prg.TrackWinVisible=YES;
  fl_show_form(fd_fTracks->fTracks,FL_PLACE_CENTERFREE,FL_FULLBORDER,"Audio Tracks");
  if(Opt->Prg.ProjectFileName[0]==0)
    fl_set_form_title(fd_fTracks->fTracks,"UNTITLED");
  else
    fl_set_form_title(fd_fTracks->fTracks,(char *)&Opt->Prg.ProjectFileName);
  fl_add_canvas_handler(fd_fTracks->canTrack,Expose,TrackWinExposeEvent,NULL);
  fl_add_canvas_handler(fd_fTracks->canTrack,ButtonPress,TrackWinButtonEvent,NULL);
  RepaintTrackWin();
}

void HideTrackWin()
{
  if(Opt->Prg.TrackWinVisible==NO)
    return;
  Opt->Prg.TrackWinVisible=NO;
//  fl_remove_canvas_handler(fd_fTracks->canTrack,Expose,NULL);
//  fl_remove_canvas_handler(fd_fTracks->canTrack,ButtonPress,NULL);
  fl_hide_form(fd_fTracks->fTracks);
  Opt->Tracks.Win=0;
  Opt->Tracks.WinWidth=0;
  Opt->Tracks.WinHeight=0;
}


int TrackWinExposeEvent(FL_OBJECT *Obj, Window win, int w, int h, XEvent *ev, void *d)
{
  Opt->Tracks.Win=win;
  Opt->Tracks.WinWidth=w;
  Opt->Tracks.WinHeight=h;
  RepaintTrackWin();
  return 0;
}


int TrackWinButtonEvent(FL_OBJECT *Obj, Window win, int w, int h, XEvent *ev, void *d)
{
  struct tAudioTrack *tmp=TrackList;
  int TrackNr=0;
  int TrH,TrW;
  int changed=0;
  TrH=Opt->Tracks.Height;
  TrW=Opt->Tracks.Width=Opt->Tracks.WinWidth-XSKIPEND-XSKIP;

  while(tmp!=NULL)
  {
    /* check action button */
    if(IsInRectangle(ev->xbutton.x,ev->xbutton.y,
                   XSKIP-20,
                   TrH*(0.05+TrackNr),
                   XSKIP-20+15,TrH*(0.05+TrackNr+0.90)))
    {
      switch(tmp->Action)
      {
        case MUTE: tmp->Action=PLAY; break;
        case PLAY: tmp->Action=REC; break;
        case REC:  tmp->Action=MUTE;
      }
      changed++;
    } /* END while is in rect */


    /* check track selection */
//    fl_rect(XSKIP,TrH*(0.05+TrackNr),TrW,TrH*0.90,FL_RED);
    if(IsInRectangle(ev->xbutton.x,ev->xbutton.y,
                   XSKIP,
                   TrH*(0.05+TrackNr),
                   XSKIP+TrW,TrH*(0.90+TrackNr)))
    {
      switch(tmp->Selected)
      {
        case NO : tmp->Selected=YES; break;
        case YES: tmp->Selected=NO; break;
      }
      changed++;
    } /* END while is in rect */
    

    tmp=tmp->next;
    TrackNr++;
  } /* END while all tracks */
  if(changed!=0)
    RepaintTrackWin();
  return 0;
}


void RepaintTrackWin()
{
  struct tAudioTrack *tmp=TrackList;
  int TrackNr=0;
  char Nname[300];
  int i,j;
  int TrH,TrW;
  int PeakFileLenght, Fak;
//  char y1l,y1r,y2l,y2r;

  char plmax,prmax;
  int  dlmax,drmax;
    
  int ly1,ly2;
  char *Peak;
  long PeakCount;

  


  if((Opt->Prg.TrackWinVisible==NO)||(Opt->Tracks.Win==0))     /* bad, but faster than fl_get_canvas_id() */
    return;
  fl_winset(Opt->Tracks.Win);   /* maybe useless ... */

  /* draw the background */
  fl_rectf(0,0,Opt->Tracks.WinWidth,Opt->Tracks.WinHeight,FL_BLACK);

  if(Opt->Tracks.NrOfTracks==0)
    TrH=100;
  else
    TrH=Opt->Tracks.WinHeight/Opt->Tracks.NrOfTracks;
  Opt->Tracks.Height=TrH;

  
  TrW=Opt->Tracks.Width=Opt->Tracks.WinWidth-XSKIPEND-XSKIP;

  /* now, draw the different tracks */
  while(tmp!=NULL)
  {
    if(TrackNr*TrH > Opt->Tracks.WinHeight)  /* anything more to draw ? */
      break;
    /* the border */
    if(tmp->Selected==NO)
      fl_rect(XSKIP,TrH*(0.05+TrackNr),TrW,TrH*0.90,FL_RED);
    else
      fl_rectf(XSKIP,TrH*(0.05+TrackNr),TrW,TrH*0.90,FL_RED);
 
    /* the blue line */
    fl_line(0,TrH+TrackNr*TrH,Opt->Tracks.WinWidth,TrH+TrackNr*TrH,FL_BLUE);
    /* the track name */
    sprintf((char *)&Nname,"%s",tmp->Name);
    fl_drw_text(FL_ALIGN_LEFT,3,TrH*TrackNr+10,0,0,FL_WHITE,
                FL_FIXEDBOLD_STYLE,FL_NORMAL_SIZE,(char *)&Nname);
    /* the track action mode (colored rect) */
    switch(tmp->Action)
    {
      case MUTE:
      {
        i=FL_YELLOW;
        sprintf((char *)&Nname,"%s","M\nU\nT\nE");
      } break;
      case PLAY:
      {
        i=FL_GREEN;
        sprintf((char *)&Nname,"%s","P\nL\nA\nY");
      } break;
      case REC:
      {
        i=FL_RED;
        sprintf((char *)&Nname,"%s","R\nE\nC");
      }
    }
    fl_rectf(XSKIP-20,TrH*(0.05+TrackNr),15,TrH*0.90,i);
    fl_drw_text(FL_ALIGN_CENTER,
                (XSKIP-20)+15/2,
                TrH*(0.05+TrackNr)+(TrH*0.90/2),
                0,0,FL_BLACK,
                FL_FIXEDBOLD_STYLE,FL_NORMAL_SIZE,(char *)&Nname);
    
    /* the sample line(s) */
    PeakFileLenght=GetMaxPeakSampleLenght();
    Fak=(PeakFileLenght/TrW)+1;
    PeakCount=0;
    Peak=tmp->PeakSamples;

    if(Peak!=NULL)
    {
      for(i=1;i<TrW;i++)
      {
        if(Opt->Audio.Channels==1)
        {
          plmax=0;

          for(j=0;j<Fak;j++)
          {
            if(PeakCount < tmp->PeakSampleLenght-1)
            {
              if(abs(*(Peak++)) > abs(plmax))
                plmax=abs(*Peak);
              PeakCount++;
            }
          }
          ly1=TrH*(TrackNr +1)-(TrH/2);      // +        (y1l*TrH)/513;
          dlmax=(plmax*TrH)/257;
          fl_line(XSKIP+i,ly1+dlmax,XSKIP+i,ly1-dlmax,FL_GREEN);

        }
        else  // 2 channels
        {
          plmax=prmax=0;

          if(Fak<2)
            Fak=2;

          for(j=0;j<Fak/2;j++)
          {
            if(PeakCount < tmp->PeakSampleLenght-2)
            {
              if(abs(*(Peak++)) > abs(plmax))
                plmax=abs(*Peak);
              if(abs(*(Peak++)) > abs(prmax))
                prmax=abs(*Peak);
              PeakCount+=2;
            }
          }
          ly1=TrH*(TrackNr +1)-3*(TrH/4);      // +        (y1l*TrH)/513;
          ly2=TrH*(TrackNr +1)-(TrH/4);  // +        (y1r*TrH)/513;
          dlmax=(plmax*TrH)/513;
          drmax=(prmax*TrH)/513;
          
          fl_line(XSKIP+i,ly1+dlmax,XSKIP+i,ly1-dlmax,FL_GREEN);
          fl_line(XSKIP+i,ly2+drmax,XSKIP+i,ly2-drmax,FL_GREEN);

        }  // END else draw 16 BIT
      }  // END for the whole screen
    }  //END if Peaklist available

    /* the file name */
    sprintf((char *)&Nname,"%s",tmp->FileName);
    fl_drw_text(FL_ALIGN_LEFT,XSKIP+5,TrH*TrackNr+15,0,0,FL_WHITE,
                FL_FIXEDBOLD_STYLE,FL_SMALL_SIZE,(char *)&Nname);

    tmp=tmp->next;
    TrackNr++;
  }  // END while !all tracks drawn
}



int IsInRectangle(int thisx, int thisy, int x1, int y1, int x2, int y2)
{
  if( (thisx > x1)&&(thisx < x2) &&
      (thisy > y1)&&(thisy < y2))
    return 1;
  else
    return 0;
}



void ShowWaitWin(char *Message)
{
  fl_show_form(fd_fStop->fStop,FL_PLACE_CENTERFREE,FL_FULLBORDER,"in action");
  fl_add_canvas_handler(fd_fStop->canStop,Expose,WaitWinExposeEvent,NULL);
//  fl_add_canvas_handler(fd_fStop->canStop,ButtonPress,WaitWinButtonEvent,NULL);
  sprintf((char *)&Opt->Prg.WaitWinMsg,"%s",Message);
  Opt->Prg.WaitWin=fl_get_canvas_id(fd_fStop->canStop);
  RepaintWaitWin();
}

void HideWaitWin()
{
  fl_remove_canvas_handler(fd_fStop->canStop,Expose,NULL);
  fl_hide_form(fd_fStop->fStop);
  Opt->Prg.WaitWin=0;
  Opt->Prg.WaitWinWidth=0;
  Opt->Prg.WaitWinHeight=0;
  Opt->Prg.WaitWinProgress=0;
  Opt->Prg.WaitWinVisible=NO;
}

int WaitWinButtonEvent(FL_OBJECT *Obj, Window win, int w, int h, XEvent *ev, void *d)
{
  RepaintWaitWin();  
  return 0;
}

int WaitWinExposeEvent(FL_OBJECT *Obj, Window win, int w, int h, XEvent *ev, void *d)
{
  Opt->Prg.WaitWin=win;
  Opt->Prg.WaitWinWidth=w;
  Opt->Prg.WaitWinHeight=h;
  Opt->Prg.WaitWinProgress=0;
  Opt->Prg.WaitWinVisible=YES;
  RepaintWaitWin();  

  return 0;
}

int WaitWinVisible()
{
  if((Opt->Prg.WaitWinWidth==0)||(Opt->Prg.WaitWinHeight==0))
    return NO;
  else
    return YES;
}

void RepaintWaitWin()
{
  float Prog=Opt->Prg.WaitWinProgress;
  int i;

  if(Opt->Prg.WaitWin==0)
    return;
  

  /* this is a terrible hack;
   * the window can not be drawn (well exactly: seen) if ExposeEvent
   * is not called first */
  while(WaitWinVisible()==NO)
  {
    fl_check_forms();
    fl_set_form_size(fd_fStop->fStop,210,131);
    fl_set_form_size(fd_fStop->fStop,210,130);
  }

  fl_winset(Opt->Prg.WaitWin);   /* maybe useless ... */
  fl_rectf(0,0,Opt->Prg.WaitWinWidth,Opt->Prg.WaitWinHeight,FL_BLACK);
  fl_drw_text(FL_ALIGN_CENTER,Opt->Prg.WaitWinWidth/2,Opt->Prg.WaitWinHeight/4 ,0,0,FL_WHITE,
                FL_FIXEDBOLD_STYLE,FL_NORMAL_SIZE,(char *)&Opt->Prg.WaitWinMsg);
  fl_rect(Opt->Prg.WaitWinWidth*0.05,
          Opt->Prg.WaitWinHeight/2,
          Opt->Prg.WaitWinWidth*0.9,
          Opt->Prg.WaitWinHeight/4,FL_GREEN);
  fl_rectf(Opt->Prg.WaitWinWidth*0.05,
          Opt->Prg.WaitWinHeight/2,
          (float)((float)Opt->Prg.WaitWinWidth*0.9)*(float)(Prog/100),
          Opt->Prg.WaitWinHeight/4,FL_RED);


  /* this is a terrible hack;
   * the form is not drawn fast enough... */

/*
  for(i=0;i<5;i++)
  {
    fl_raise_form(fd_fStop->fStop);
    fl_check_forms();
  }
*/
  fl_check_forms();
}

void SetWaitWinProgress(int Percent)
{
  if(Opt->Prg.WaitWin==0)
  {
    Opt->Prg.WaitWinProgress=0;
    return;
  }

  if(Percent < 0)
    Percent=0;
  if(Percent > 100)
    Percent=100;
  
  Opt->Prg.WaitWinProgress=Percent;
  RepaintWaitWin();
}


void ShowStopWin()
{
  fl_show_form(fd_fStop->fStop,FL_PLACE_CENTERFREE,FL_FULLBORDER,"Audio Tracks");
  fl_add_canvas_handler(fd_fStop->canStop,Expose,StopWinExposeEvent,NULL);
  fl_add_canvas_handler(fd_fStop->canStop,ButtonPress,StopWinButtonEvent,NULL);
  fl_set_form_size(fd_fStop->fStop,210,131);
  fl_set_form_size(fd_fStop->fStop,210,130);
}

void HideStopWin()
{
  fl_remove_canvas_handler(fd_fStop->canStop,Expose,NULL);
  fl_remove_canvas_handler(fd_fStop->canStop,ButtonPress,NULL);
  fl_hide_form(fd_fStop->fStop);
}


int StopWinExposeEvent(FL_OBJECT *Obj, Window win, int w, int h, XEvent *ev, void *d)
{
  fl_winset(win); 
  fl_rectf(0,0,w,h,FL_BLACK);
  fl_rectf((w/2)-h/4,(h/2)-(h/4),h/2,h/2,FL_WHITE);

  return 0;
}


int StopWinButtonEvent(FL_OBJECT *Obj, Window win, int w, int h, XEvent *ev, void *d)
{
#ifdef DEBUG
  fprintf(stderr,"STOP\n");
#endif
  Opt->Prg.StopAction=YES;
  return 0;
}











/************************************************************
 *** NotImplemented
 *** shows a window that this function is not available yet
 ************************************************************/
void NotImplemented()
{
  fl_show_messages("this function has not been implemented yet");
}









