#include "forms.h"
#include "xwin2.h"
#include "global.h"
#include "windows.h"
#include "file.h"
#include "error.h"
#include "tracks.h"
#include "audio.h"
#include "tools.h"


extern FD_fMenu *fd_fMenu;
extern FD_fSetPrg *fd_fSetPrg;
extern FD_fSetAudio *fd_fSetAudio;
extern FD_fTracks *fd_fTracks;
extern FD_fMeter *fd_fMeter;



/* callbacks and freeobj handles for form fMenu */
void cb_mProject(FL_OBJECT *ob, long data)
{
  char *FileName;
  FILE *PrjFileDummy;

  switch(fl_get_menu(ob))
  {

    /* NEW */
    case 1:
    {
      if(TrackList!=NULL)
      {
        if(fl_show_question("really create a new project?\nall current tracks will be removed!",2)!=0)
        {
          RemoveAllTracksFromList();
          sprintf((char *)&Opt->Prg.ProjectFileName,"");
          fl_set_form_title(fd_fTracks->fTracks,"UNTITLED");
          RepaintTrackWin();
        }
      }
    } break;

    /* OPEN */
    case 2:
    {
      if(TrackList!=NULL)
      {
        if(fl_show_question("really open a new project?\nall current tracks will be removed!",2)==0)
          break;
      }

      FileName=(char *)fl_show_fselector("",(char *)&Opt->Prg.AudioDir,"*.xhdprj","");
      if(FileName!=NULL)
      {
        RemoveAllTracksFromList();
        CutLF(FileName,299);
        ShowWaitWin("loading project...");
        SetWaitWinProgress(0);
        if(ReadProjectFile(FileName)!=0)
        {
          ShowErrorMsg();
          sprintf((char *)&Opt->Prg.ProjectFileName,"");
          fl_set_form_title(fd_fTracks->fTracks,"UNTITLED");
        }
        else
          fl_set_form_title(fd_fTracks->fTracks,Opt->Prg.ProjectFileName);

        HideWaitWin();
      }
    } break;
    
    /* SAVE */
    case 3:
    {
      if(TrackList==NULL)
      {
        fl_show_messages("nothing to save");
        break;
      }

      if(Opt->Prg.ProjectFileName[0]==0)
      {
        FileName=(char *)fl_show_fselector("",(char *)&Opt->Prg.AudioDir,"*.xhdprj","");
        if(FileName==NULL)
          break;

        AddExtensionString(FileName,".xhdprj");

        if((PrjFileDummy=OpenFileForPlay(FileName))!=NULL)
        {
          CloseFile(PrjFileDummy);
          if(fl_show_question("file exists! overwrite?",2)==0)
            break;
        }
      }
      else
        FileName=(char *)&Opt->Prg.ProjectFileName;
        
      if(FileName!=NULL)
      {
        CutLF(FileName,299);
        if(CreateProjectFile(FileName)!=0)
          ShowErrorMsg();
        else
        {
          sprintf((char *)&Opt->Prg.ProjectFileName,"%s",FileName);
          fl_set_form_title(fd_fTracks->fTracks,Opt->Prg.ProjectFileName);
        }
      }
    } break;

    /* SAVE AS */
    case 4:
    {
      if(TrackList==NULL)
      {
        fl_show_messages("nothing to save");
        break;
      }

      FileName=(char *)fl_show_fselector("",(char *)&Opt->Prg.AudioDir,"*.xhdprj","");
      if(FileName==NULL)
        break;

      AddExtensionString(FileName,".xhdprj");

#ifdef DEBUG
      fprintf(stderr,"FILE TO SAVE: %s",FileName);
#endif
      

      if((PrjFileDummy=OpenFileForPlay(FileName))!=NULL)
      {
        CloseFile(PrjFileDummy);
        if(fl_show_question("file exists! overwrite?",2)==0)
          break;
      }

      if(FileName!=NULL)
      {
        CutLF(FileName,299);
        if(CreateProjectFile(FileName)!=0)
          ShowErrorMsg();
        else
        {
          sprintf((char *)&Opt->Prg.ProjectFileName,"%s",FileName);
          fl_set_form_title(fd_fTracks->fTracks,Opt->Prg.ProjectFileName);
        }
      }
    } break;

    /* QUIT */
    case 5:
    {
      if(fl_show_question("Really Quit ?",2)!=0)
      {
        RemoveAllTracksFromList();
        fl_finish();
        exit(0);
      }
    } break;
    
  }
}


void cb_mSettings(FL_OBJECT *ob, long data)
{
  switch(fl_get_menu(ob))
  {
    /* PROGRAM-SETTINGS-MENU */
    case 1:
    {
      fl_show_form(fd_fSetPrg->fSetPrg,FL_PLACE_MOUSE,FL_FULLBORDER,"fMenu");
    } break;

    /* AUDIO_SETTINGS_MENU */
    case 2:
    {
      fl_show_form(fd_fSetAudio->fSetAudio,FL_PLACE_MOUSE,FL_FULLBORDER,"fMenu");
    } break;

    /* SAVE SETTINGS */
    case 3:
    {
      ReadProgramSettingsWin();
      ReadAudioSettingsWin();
      if(CreateConfigFile()==0)
        fl_show_messages("options saved");
      else
        ShowErrorMsg();

    }
  }
} 

void cb_mShow(FL_OBJECT *ob, long data)
{
  switch(fl_get_menu(ob))
  {
    /* VU_METER */
    case 1:
    {
      if(Opt->Prg.VUMeterWinVisible==YES)
      {
        Opt->Prg.VUMeterWinVisible=NO;
        fl_hide_form(fd_fMeter->fMeter);
      }	
      else
      {
        Opt->Prg.VUMeterWinVisible=YES;
	fl_show_form(fd_fMeter->fMeter,FL_PLACE_CENTER,FL_FULLBORDER,"VU Meter");
      }
    } break;
    
    /* TRACKS */
    case 2:
    {
      if(Opt->Prg.TrackWinVisible==NO)
        ShowTrackWin();
      else
        HideTrackWin();
    } break;
  }
}

void cb_mTrack(FL_OBJECT *ob, long data)
{

  switch(fl_get_menu(ob))
  {
    /* ADD */
    case 1:
    {
      AddTrackToProject();
      
    } break;

    /* REMOVE */
    case 2:
    {
      RemoveTrackFromProject();
    } break;
    
    /* EDIT */
    case 3:
    {
    } break;

    /* IMPORT */
    case 4:
    {
      ImportTrackToProject();
    } break;

    /* EXPORT */
    case 5:
    {
    } break;
  }
}



/* callbacks and freeobj handles for form fSetPrg */
void cb_iPrgOptAudioDir(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}

void cb_butPrgOptOK(FL_OBJECT *ob, long data)
{
  ReadProgramSettingsWin();
  fl_hide_form(fd_fSetPrg->fSetPrg);
}

void cb_butPrgOptCancel(FL_OBJECT *ob, long data)
{
  UpdateProgramSettingsWin();
  fl_hide_form(fd_fSetPrg->fSetPrg);
}

void cb_iPrgOptMixerApp(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}

void cb_butPrgOptBrowseAudioDir(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}

void cb_butPrgOptBrowseMixerApp(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}

void cb_chPrgOptHideWindows(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}

void cb_chPrgOptRecCount(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}



/* callbacks and freeobj handles for form fSetAudio */
void cb_iAudOptPlaybackDev(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}

void cb_iAudOptRecordDev(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}

void cb_butAudOptOK(FL_OBJECT *ob, long data)
{
  ReadAudioSettingsWin();
  fl_hide_form(fd_fSetAudio->fSetAudio);
}

void cb_butAudOptCancel(FL_OBJECT *ob, long data)
{
  UpdateAudioSettingsWin();
  fl_hide_form(fd_fSetAudio->fSetAudio);
}

void cb_chAudOptFullDuplex(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}


void cb_butAudOptTestPlay(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}

void cb_butAudOptTestRecord(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}

void cb_chAudOptBits(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}

void cb_chAudOptChannels(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}

void cb_chAudOptSampleRate(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}

void cb_chAudOptRecMode(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}

void cb_chAudOptPlayMode(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}


/* added manually */
void cb_iAudOptPlayFragsize(FL_OBJECT *ob, long data)
{
  Opt->Audio.PlayInfoTested=YES;
}
void cb_iAudOptPlayFragments(FL_OBJECT *ob, long data)
{
  Opt->Audio.PlayInfoTested=YES;
}
void cb_butAudOptGetPlayInfo(FL_OBJECT *ob, long data)
{
  ReadAudioSettingsWin();
  if(GetBestPlayInfo()!=0)
    ShowErrorMsg();
  UpdateAudioSettingsWin();
}
void cb_butAudOptTestPlayInfo(FL_OBJECT *ob, long data)
{
  ReadAudioSettingsWin();
  if(TestPlayInfo()!=0)
    ShowErrorMsg();
  UpdateAudioSettingsWin();
}




void cb_iAudOptRecFragsize(FL_OBJECT *ob, long data)
{
  Opt->Audio.RecInfoTested=YES;
}
void cb_iAudOptRecFragments(FL_OBJECT *ob, long data)
{
  Opt->Audio.RecInfoTested=YES;
}
void cb_butAudOptGetRecInfo(FL_OBJECT *ob, long data)
{
  ReadAudioSettingsWin();
  if(GetBestRecInfo()!=0)
    ShowErrorMsg();
  UpdateAudioSettingsWin();
}
void cb_butAudOptTestRecInfo(FL_OBJECT *ob, long data)
{
  ReadAudioSettingsWin();
  if(TestRecInfo()!=0)
    ShowErrorMsg();
  UpdateAudioSettingsWin();
}





void cb_iAudOptDuplexPlayFragsize(FL_OBJECT *ob, long data)
{
  Opt->Audio.DuplexInfoTested=YES;
}
void cb_iAudOptDuplexPlayFragments(FL_OBJECT *ob, long data)
{
  Opt->Audio.DuplexInfoTested=YES;
}
void cb_iAudOptDuplexRecFragsize(FL_OBJECT *ob, long data)
{
  Opt->Audio.DuplexInfoTested=YES;
}
void cb_iAudOptDuplexRecFragments(FL_OBJECT *ob, long data)
{
  Opt->Audio.DuplexInfoTested=YES;
}
void cb_butAudOptGetDuplexInfo(FL_OBJECT *ob, long data)
{
  ReadAudioSettingsWin();
  if(Opt->Audio.FullDuplex==YES)
  {
    if(GetBestDuplexInfo()!=0)
      ShowErrorMsg();
  }
  else
    fl_show_messages("this option requires full duplex mode!");
  UpdateAudioSettingsWin();
}
void cb_butAudOptTestDuplexInfo(FL_OBJECT *ob, long data)
{
  
  ReadAudioSettingsWin();
  if(Opt->Audio.FullDuplex==YES)
  {
    if(TestDuplexInfo()!=0)
      ShowErrorMsg();
  }
  else
    fl_show_messages("this option requires full duplex mode!");
  UpdateAudioSettingsWin();
}




/* callbacks and freeobj handles for form fTracks */
void cb_canTrack(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}

void cb_sliTrackZoom(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}

void cb_sliTrackScroll(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}

void cb_butTrackPlay(FL_OBJECT *ob, long data)
{
  int TrackMode;
  int DeviceMode;


  switch((TrackMode=GetNecessaryPlayMode())) 
  {
    case MUTE        : return;
    case PLAY        :
    {
      if(Opt->Audio.PlayInfoTested==NO)
      {
        fl_show_messages("play fragments and -size not tested!\nplease set audio options!");
        return;
      }
      DeviceMode=PLAYTRACKS;
    } break;
    case REC         :
    {
      if(Opt->Audio.RecInfoTested==NO)
      {
        fl_show_messages("rec fragments and -size not tested!\nplease set audio options!");
        return;
      }
      DeviceMode=RECTRACKS;
    } break;
    case (PLAY | REC):
    {
      if(Opt->Audio.DuplexInfoTested==NO)
      {
        fl_show_messages("duplex fragments and -size not tested!\nplease set audio options!");
        return;
      }
      DeviceMode=PLAYANDRECTRACKS;
    } break;
    default          : return;
  }


  if(InitSoundDevice(DeviceMode)==0)
  {

    /* TODO */
    HideTrackWin();
    HideMenuWin();
    ShowStopWin();
    Opt->Prg.StopAction=NO;

    switch(TrackMode) 
    {
      case PLAY: 
      {
      
        if(PlayMultipleTracks2()!=0)
          ShowErrorMsg();
      } break;

      case REC:
      {
        if(RecMultipleTracks2()!=0)
          ShowErrorMsg();
      } break;

      case (PLAY | REC):
      {
        if(PlayAndRecMultipleTracks2()!=0)
          ShowErrorMsg();
        break; 
      }
    } // END switch PlayMode

    ShowMenuWin();
    ShowTrackWin();
    HideStopWin();
    /* looks strange, but is needed due to a bug in xforms */
    HideTrackWin();
    ShowTrackWin();
    Opt->Prg.StopAction=NO;

    if(CloseSoundDevice()!=0)
      ShowErrorMsg();
  } // END if InitSoundDevice()
  else
  {
    ShowErrorMsg();
    if(CloseSoundDevice()!=0)
      ShowErrorMsg();
  }
}

void cb_butTrackStop(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}

void cb_butTrackRec(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}

void cb_butTrackClose(FL_OBJECT *ob, long data)
{
  HideTrackWin();
}



/* callbacks and freeobj handles for form fMeter */

/* callbacks and freeobj handles for form fStop */
void cb_canStop(FL_OBJECT *ob, long data)
{
  /* fill-in code for callback */
}


